@@ -52,10 +52,7 @@
 #include "global.h"
 
 #include <stdio.h>
-#ifdef HAVE_READLINE
-# include <readline/readline.h>
-# include <readline/history.h>
-#endif
+#include <sys/stat.h>
 #include <stdlib.h>
 #include <unistd.h>
 #include <string.h>
@@ -67,6 +64,7 @@
 #include <X11/Xlib.h>
 
 #include "hp48.h"
+#include "hp48_emu.h"
 #include "device.h"
 #include "timer.h"
 #include "x48_x11.h"
@@ -75,6 +73,10 @@
 #include "rpl.h"
 #include "romio.h"
 #include "resources.h"
+#include "init.h"
+
+extern char *readline __ProtoType__ ((const char *));
+extern void add_history __ProtoType__ ((char *));
 
 #define MAX_ARGS 16
 
@@ -90,9 +92,6 @@ static char instr[100];
  */
 #define DSKTOP_SX	0x70579
 #define DSKBOT_SX	0x7057e
-/*
- * Pointers in the HP48gx ROM
- */
 #define DSKTOP_GX	0x806f8
 #define DSKBOT_GX	0x806fd
 
@@ -121,20 +120,29 @@ bkpt_tbl[MAX_BREAKPOINTS + 1];
  */
 static void do_break __ProtoType__ ((int, char **));
 static void do_continue __ProtoType__ ((int, char **));
+static void do_crc __ProtoType__ ((int, char **));
 static void do_delete __ProtoType__ ((int, char **));
+static void do_dis __ProtoType__ ((int, char **));
+static void do_diso __ProtoType__ ((int, char **));
 static void do_exit __ProtoType__ ((int, char **));
+static void do_find __ProtoType__ ((int, char **));
 static void do_go __ProtoType__ ((int, char **));
 static void do_help __ProtoType__ ((int, char **));
 static void do_load __ProtoType__ ((int, char **));
+static void do_list __ProtoType__ ((int, char **));
+static void do_ls __ProtoType__ ((int, char **));
 static void do_mode __ProtoType__ ((int, char **));
 static void do_quit __ProtoType__ ((int, char **));
 static void do_regs __ProtoType__ ((int, char **));
+static void do_receive __ProtoType__ ((int, char **));
 static void do_save __ProtoType__ ((int, char **));
+static void do_send __ProtoType__ ((int, char **));
 static void do_stack __ProtoType__ ((int, char **));
 static void do_stat __ProtoType__ ((int, char **));
 static void do_step __ProtoType__ ((int, char **));
+static void do_trace __ProtoType__ ((int, char **));
 static void do_ram __ProtoType__ ((int, char **));
-static void do_reset __ProtoType__ ((int, char **));
+static void do_reset_calc __ProtoType__ ((int, char **));
 static void do_rstk __ProtoType__ ((int, char **));
 
 struct cmd
@@ -148,7 +156,7 @@ cmd_tbl[] =
 {
   {
     "break", do_break,
-    "break [address]            Set breakpoint at `address\' or show breakpoints"
+    "break [address ]            Set breakpoint at `address\' or show breakpoints,\n      [addr_low addr_high]  Set breakpoint range, from low to high,\n      [sysRPL_keyword]      Set breakpoint at System RPL keyword "
   }
   ,
   {
@@ -161,26 +169,40 @@ cmd_tbl[] =
     "cont                       Continue execution"
   }
   ,
+  {
+    "crc", do_crc,
+    "crc addr len               Calc crc of memory at addr to len"
+  }
+  ,
   {
     "c", do_continue, 0
   }
   ,
-
+  {
+    "dis", do_dis,
+    "dis [addr] [addr n]        Disassemble starting at PC or addr, n times"
+  }
+  ,
+  {
+    "diso", do_diso,
+    "diso [addr]                Disassemble object at addr"
+  }
+  ,
   {
     "delete", do_delete,
     "delete [all | n]           Delete breakpoint or watchpoint number `n\',\n                           all breakpoints, or current breakpoint"
   }
   ,
-  {
-    "d", do_delete, 0
-  }
-  ,
-
   {
     "exit", do_exit,
     "exit                       Exit the emulator without saving"
   }
   ,
+  {
+    "find", do_find,
+    "find addr obj              Search for obj starting at addr"
+  }
+  ,
 
   {
     "go", do_go,
@@ -201,16 +223,25 @@ cmd_tbl[] =
     "?", do_help, 0
   }
   ,
-
+  {
+    "list", do_list,
+    "list [addr] [addr n]       List data starting at addr, n times "
+  }
+  ,
   {
     "load", do_load,
     "load                       Load emulator-state from files"
   }
   ,
+  {
+    "ls", do_ls,
+    "ls addr                    list directory contents"
+  }
+  ,
 
   {
     "mode", do_mode,
-    "mode [hp | class]          Show or set disassembler mode"
+    "mode [hp | class | sysRPL]          Show or set disassembler mode"
   }
   ,
 
@@ -239,9 +270,13 @@ cmd_tbl[] =
     "r", do_regs, 0
   }
   ,
-
   {
-    "reset", do_reset,
+    "receive", do_receive,
+    "receive file               receive file into HP48"
+  }
+  ,
+  {
+    "reset", do_reset_calc,
     "reset                      Set the HP48\'s PC to ZERO"
   }
   ,
@@ -251,7 +286,11 @@ cmd_tbl[] =
     "save                       Save emulator-state to files"
   }
   ,
-
+  {
+    "send", do_send,
+    "send ob                  Send ob to file"
+  }
+  ,
   {
     "stack", do_stack,
     "stack                      Display RPL stack"
@@ -273,6 +312,10 @@ cmd_tbl[] =
     "s", do_step, 0
   }
   ,
+  {
+   "trace", do_trace, 0
+  }
+  ,
 
   {
     "where", do_rstk,
@@ -348,6 +391,7 @@ check_breakpoint (type, addr)
   return 0;
 }
 
+
 char *
 #ifdef __FunctionProto__
 read_str(char *str, int n, int fp)
@@ -365,10 +409,7 @@ read_str(str, n, fp)
     {
       cc = read(fp, str, n);
       if (cc > 0)
-        {
-	  str[cc] = '\0';
-          return str;
-	}
+        return str;
       if (cc == 0)
         return NULL;
 
@@ -382,52 +423,11 @@ read_str(str, n, fp)
           fcntl(fp, F_SETFL, flags);
           continue;
         }
+ 
       return NULL;
     }
-    /* not reached */
 }
 
-#ifndef HAVE_READLINE
-char *
-#ifdef __FunctionProto__
-readline(char *prompt)
-#else
-readline(prompt)
-	char *prompt;
-#endif
-{
-  char rl[81];
-  char *nrl;
-  if (prompt)
-    {
-      printf("%s", prompt);
-      fflush(stdout);
-    }
-  nrl = read_str(rl, 80, 0);
-  if (nrl)
-    {
-      int len;
-      char *str;
-
-      len = strlen(nrl);
-      str = malloc(len + 1);
-
-      if (nrl[len-1] == '\n')
-        {
-	  nrl[len-1] = '\0';
-	}
-      if (NULL == str)
-        {
-	  fprintf(stderr, "Out of memory\n");
-	  exit(-1);
-        }
-
-      strcpy(str, nrl);
-      nrl = str;
-    }
-  return nrl;
-}
-#endif
 
 static inline void
 #ifdef __FunctionProto__
@@ -493,13 +493,14 @@ decode_20 (addr, arg)
 #endif
 {
   int i;
-
+  
   if (arg == (char *) 0)
     {
       printf ("Command requires an argument.\n");
       return 0;
     }
 
+
   *addr = 0;
   for (i = 0; i < strlen (arg); i++)
     {
@@ -533,6 +534,7 @@ decode_32 (addr, arg)
 #endif
 {
   int i;
+  int sign;
 
   if (arg == (char *) 0)
     {
@@ -541,24 +543,31 @@ decode_32 (addr, arg)
     }
 
   *addr = 0;
+  sign = 1;
   for (i = 0; i < strlen (arg); i++)
     {
-      *addr <<= 4;
-      if ('0' <= arg[i] && arg[i] <= '9')
-	{
-	  *addr |= ((int) arg[i] - (int) '0');
-	}
-      else if ('A' <= arg[i] && arg[i] <= 'F')
-	{
-	  *addr |= ((int) arg[i] - (int) 'A' + 10);
-	}
+      if ( i == 0 && arg[0] == (int) '-' )
+        sign = -1;
       else
-	{
-	  *addr = 0;
-	  printf ("Not a number: %s.\n", arg);
-	  return 0;
-	}
+      {
+        *addr <<= 4;
+        if ('0' <= arg[i] && arg[i] <= '9')
+    	  {
+	    *addr |= ((int) arg[i] - (int) '0');
+  	  }
+        else if ('A' <= arg[i] && arg[i] <= 'F')
+	  {
+	    *addr |= ((int) arg[i] - (int) 'A' + 10);
+	  }
+        else
+	  {
+	    *addr = 0;
+	    printf ("Not a number: %s.\n", arg);
+	    return 0;
+	  }
+       }
     }
+  *addr = *addr * sign;
   return 1;
 }
 
@@ -578,27 +587,7 @@ decode_64 (addr, arg)
       printf ("Command requires an argument.\n");
       return 0;
     }
-#ifdef SIMPLE_64
-  *addr = 0;
-  for (i = 0; i < strlen (arg); i++)
-    {
-      *addr <<= 4;
-      if ('0' <= arg[i] && arg[i] <= '9')
-	{
-	  *addr |= ((int) arg[i] - (int) '0');
-	}
-      else if ('A' <= arg[i] && arg[i] <= 'F')
-	{
-	  *addr |= ((int) arg[i] - (int) 'A' + 10);
-	}
-      else
-	{
-	  *addr = 0;
-	  printf ("Not a number: %s.\n", arg);
-	  return 0;
-	}
-    }
-#else
+
   addr->lo = addr->hi = 0;
   for (i = 0; i < strlen (arg); i++)
     {
@@ -620,7 +609,6 @@ decode_64 (addr, arg)
 	  return 0;
 	}
     }
-#endif
   return 1;
 }
 
@@ -693,18 +681,29 @@ do_break (argc, argv)
      char *argv;
 #endif
 {
-  int i;
-  word_20 addr;
+  int i,j;
+  word_20 addr, addrl, addrh;
+  char in[20];
 
-  if (argc == 1)
+  switch(argc) 
     {
-      for (i = 0; i < MAX_BREAKPOINTS; i++)
+      case 1:
+        for (i = 0; i < MAX_BREAKPOINTS; i++)
 	{
 	  if (bkpt_tbl[i].flags == 0)
 	    continue;
 	  if (bkpt_tbl[i].flags == BP_EXEC)
 	    {
-	      printf ("Breakpoint %d at 0x%.5lX\n", i + 1, bkpt_tbl[i].addr);
+	      printf ("Breakpoint %d at 0x%.5lX", i + 1, bkpt_tbl[i].addr);
+
+              j = 0;
+              while ( j < 2999 && saturn.entries[j].rom != bkpt_tbl[i].addr )
+                j++;
+              if ( j < 2990 ) 
+                printf (" : %s\n",saturn.entries[j].defn); 
+              else
+                printf ("\n");
+
 	    }
 	  else if (bkpt_tbl[i].flags == BP_RANGE)
 	    {
@@ -716,26 +715,117 @@ do_break (argc, argv)
 	      printf ("Watchpoint %d at 0x%.5lX\n", i + 1, bkpt_tbl[i].addr);
 	    }
 	}
-    }
-  else
-    {
-      str_to_upper (argv[1]);
-      if (!decode_20 (&addr, argv[1]))
-	{
-	  return;
-	}
+      break;
+  case 2:
+      i = 0;
+      strcpy( in, argv[1]);
+      strcat( in, "            ");
+      while ( i < 2999 && strncmp(saturn.entries[i].defn, in, 9 ))
+        i++;
+      if ( i < 2990 )
+          addr = saturn.entries[i].rom;
+      else
+        {
+          str_to_upper (argv[1]);
+          if (!decode_20 (&addr, argv[1]))
+       	    {
+	      return;
+	    }
+        }
       for (i = 0; i < MAX_BREAKPOINTS; i++)
 	{
 	  if (bkpt_tbl[i].flags == 0)
 	    {
 	      bkpt_tbl[i].flags = BP_EXEC;
 	      bkpt_tbl[i].addr = addr;
-	      printf ("Breakpoint %d at 0x%.5lX\n", i + 1, bkpt_tbl[i].addr);
+	      printf ("Breakpoint %d at 0x%.5lX", i + 1, bkpt_tbl[i].addr);
+ 
+              j = 0;
+              while ( j < 2999 && saturn.entries[j].rom != bkpt_tbl[i].addr )
+                j++;
+              if ( j < 2990 )        
+                printf (" : %s\n",saturn.entries[j].defn);
+              else
+                printf ("\n");
+
 	      num_bkpts++;
 	      return;
 	    }
 	}
       printf ("Breakpoint table full\n");
+     break;
+
+  case 3:
+    str_to_upper (argv[1]);
+    str_to_upper (argv[2]);
+    addr = addrh = addrl = 0;
+    decode_20 (&addrl, argv[1]);
+    decode_20 (&addrh, argv[2]);
+
+    if ( addrl < 1 && addrh < 1 )
+      {
+        return;
+      }
+    
+    if ( addrl == addrh ) 
+       return;
+
+    if (addrl > addrh )
+      {
+        addr = addrl;
+        addrl = addrh;
+        addrh = addr;
+      }
+    for (i = 0; i < MAX_BREAKPOINTS; i++)
+      {
+        if (bkpt_tbl[i].flags == 0)
+          {
+            bkpt_tbl[i].flags = BP_RANGE;
+            bkpt_tbl[i].addr = addrl;
+            bkpt_tbl[i].end_addr = addrh;
+            printf ("Range watchpoint %d at 0x%.5lX - 0x%.5lX\n", i + 1,
+                    bkpt_tbl[i].addr, bkpt_tbl[i].end_addr);
+            num_bkpts++;
+            return;
+          }
+      }
+    printf ("Breakpoint table full\n");
+    break;
+  }
+
+
+}
+
+static void
+#ifdef __FunctionProto__
+do_crc (int argc, char **argv)
+#else
+do_crc (argc, argv)
+     int argc;
+     char *argv;
+#endif
+{
+  word_20 addr, nib, crc, start, len, stop;
+
+  if (argc == 1)
+    {
+    stop = 0xC003E;
+     {
+      start = 0xC0005;
+        {
+           crc = 0;
+           addr = start;
+           len = 0;
+           while (addr != stop )
+             {
+               nib = read_nibble(addr);
+               crc = (crc >> 4) ^ (((crc ^ nib) & 0xf) * 0x1081);
+               addr++;
+               len++;
+             }
+              printf("CRC from %.5lX -> %.5lX : %lX\n", start, start+len, crc); 
+        }
+      }
     }
 }
 
@@ -751,6 +841,85 @@ do_continue (argc, argv)
   continue_flag = 1;
 }
 
+static void
+#ifdef __FunctionProto__
+do_diso (int argc, char **argv)
+#else
+do_diso (argc, argv)
+     int argc;     char *argv;
+#endif
+{
+  word_20 addr, len ;
+  char    buf[65536];
+
+
+  str_to_upper (argv[1]);
+  if (decode_20 (&addr, argv[1]))
+  {
+
+    decode_rpl_obj(addr, &len, buf);
+       if (strlen(buf) > 63)
+         {
+           sprintf(&buf[60], "...");
+           buf[63] = '\0';
+         }
+     printf("%.5lX -> %.5lX : %s\n", addr, addr+len , buf);
+
+   }
+}
+
+static void
+#ifdef __FunctionProto__
+do_dis (int argc, char **argv)
+#else
+do_dis (argc, argv)
+     int argc;     char *argv;
+#endif
+{
+  word_20 addr,naddr,val ;
+  int n;
+
+  switch (argc)
+  { 
+    case 1:
+        naddr = disassemble (saturn.PC, instr);
+        val = read_nibbles(saturn.PC, naddr - saturn.PC);
+        printf ("    PC:\t%.5lX\t%.5lX -> %s\n", saturn.PC, val, instr);
+        break;
+    case 2: 
+      str_to_upper (argv[1]);
+      if (decode_20 (&addr, argv[1]))
+      {
+        disassemble (addr, instr);
+        printf ("    PC:\t%.5lX -> %s\n", addr, instr);
+      }
+      break;
+    case 3:
+      str_to_upper (argv[1]);
+      addr = -1;
+      if (!strncmp(argv[1],"PC",2))
+          addr = saturn.PC;
+      else if (!decode_20 (&addr, argv[1]))
+        return;
+      str_to_upper (argv[2]);
+      if (decode_dec(&n, argv[2]) )
+      {
+        if (n <= 0)
+          return;
+
+        for ( ; n >= 1 ; n--)
+        {
+           naddr = disassemble (addr, instr);
+           printf ("    PC:\t%.5lX -> %s\n", addr, instr);
+           addr = naddr;
+         }
+
+      }
+      break;
+  }
+
+}
+
 static void
 #ifdef __FunctionProto__
 do_delete (int argc, char **argv)
@@ -852,6 +1021,34 @@ do_exit (argc, argv)
     }
 }
 
+static void
+#ifdef __FunctionProto__
+do_find (int argc, char **argv)
+#else
+do_find (argc, argv)
+     int argc;
+     char *argv;
+#endif
+{
+  word_32 n, addr;
+ 
+  if (argc == 3 )
+  {
+    str_to_upper (argv[1]);
+    str_to_upper (argv[2]);
+    if (decode_32 (&addr, argv[1]) && decode_32(&n, argv[2]) )
+      {
+         if (n >= 0 && addr >=0 )
+           {
+              while ( n != read_nibbles(addr, 5) && addr < 0xffffe )
+                {
+                 addr++;
+                }
+           }
+      }
+    printf ("finished at %.5lX\n", addr);
+   }
+}
 static void
 #ifdef __FunctionProto__
 do_go (int argc, char **argv)
@@ -862,13 +1059,16 @@ do_go (argc, argv)
 #endif
 {
   word_20 addr;
-
-  str_to_upper (argv[1]);
-  if (decode_20 (&addr, argv[1]))
-    {
-      saturn.PC = addr;
-      enter_debugger &= ~ILLEGAL_INSTRUCTION;
-    }
+  
+  if (argc == 2 )
+  {
+    str_to_upper (argv[1]);
+    if (decode_20 (&addr, argv[1]))
+      {
+        saturn.PC = addr;
+        enter_debugger &= ~ILLEGAL_INSTRUCTION;
+      }
+  }
 }
 
 static void
@@ -891,6 +1091,224 @@ do_help (argc, argv)
     }
 }
 
+static word_20
+#ifdef __FunctionProto__
+where (char *ls)
+#else
+where (ls)
+     char *ls;
+#endif
+{
+
+  word_20 addr,len, start;
+  unsigned char    name[64];
+  int     leave, home ,i ;
+
+  addr = read_nibbles(0x7059C, 5);
+
+  if (addr == read_nibbles(0x70592,5))
+    home = 1;
+  else
+    home =0;
+
+  if ( read_nibbles(addr, 5) == DORRP )
+    {
+      addr += 5;
+      start = read_nibbles(addr,3);
+      addr += 3;
+      if ( home )
+        {
+          for ( ; start >= 1 ; start-- )
+              addr+=13;
+        }
+
+      if ( read_nibbles(addr,5) != 0 )
+        {
+          addr += read_nibbles(addr,5);
+          start = addr-5;
+          leave = 0;
+          while (! leave ) 
+            {
+              len = read_nibbles(addr, 2);
+              addr +=2;
+              if (len !=0 )
+               {
+                 i = 0;
+                 for ( ; len >= 1; len -- )
+                   {
+                      name[i] = (int)read_nibbles(addr, 2);
+                      i++;
+                      addr +=2;
+                   }
+                 name[i]=0;
+                 addr +=2;
+               }
+             
+               if( ! strncmp(name, ls, strlen(ls) ))
+                  return addr;
+    
+              if ( read_nibbles(start,5) == 0)
+                leave = 1;
+              else
+                {
+                  addr = start - read_nibbles(start, 5);
+                  start = addr - 5;
+                }
+            } 
+
+        }
+      }
+
+  return 0;
+}
+
+static void
+#ifdef __FunctionProto__
+do_ls (int argc, char **argv)
+#else
+do_ls (argc, argv)
+     int argc;
+     char *argv;
+#endif
+{
+  word_20 addr,len, start, length;
+  char    buf[65536];
+  int     leave, home, i;
+
+  if (argc == 1)
+    addr = read_nibbles(0x7059C, 5);
+  else if (argc == 2)
+    {
+      str_to_upper (argv[1]);
+      if (!decode_20 (&addr, argv[1]))
+        return;
+    }
+  else 
+    return;
+ 
+  if (addr == read_nibbles(0x70592,5))
+    home = 1;
+  else
+    home =0;
+
+  if ( read_nibbles(addr, 5) == DORRP )
+    {
+      printf(" Directory at %.5lX\n", addr); /* DORRP */
+      addr += 5;
+
+      start = read_nibbles(addr,3);
+      addr += 3;
+      if ( home )
+        {
+          for ( ; start >= 1 ; start-- )
+            {
+              printf( "  Library attached: %i\n", (int) read_nibbles(addr,3));
+              addr+=13;
+            }
+        }
+      else
+        {
+          if ( start != 2047 ) 
+            printf("  Library Attached: %i\n", (int)start); /* ROMPARTID */
+          else
+            printf(" No Libraries attached to this directory\n");
+        }
+
+      if ( read_nibbles(addr,5) == 0 )
+        printf(" Empty directory\n");
+      else
+        {
+          addr += read_nibbles(addr,5);
+          start = addr-5;
+          leave = 0;
+          while (! leave ) 
+            {
+              len = read_nibbles(addr, 2);
+              addr +=2;
+              if (len !=0 )
+               {
+                 printf("  ");
+                 i=len;
+                 for ( ; len >= 1; len -- )
+                   {
+                      printf ("%c",(int)read_nibbles(addr, 2));
+                      addr +=2;
+                   }
+                 for (; i <=8 ; i ++)
+                    printf(" ");
+                 addr +=2;
+               }
+              else
+                printf("  NULL     " );
+              printf(" : @ %.5lX ", addr);
+              decode_rpl_obj(addr, &length,  buf);
+              printf("len: %lX ", length);
+              if (strlen(buf) > 43)
+                {
+                  sprintf(&buf[40], "...");
+                  buf[43] = '\0';
+                }
+              printf(" : %s\n", buf );
+    
+              if ( read_nibbles(start,5) == 0)
+                leave = 1;
+              else
+                {
+                  addr = start - read_nibbles(start, 5);
+                  start = addr - 5;
+                }
+
+            } 
+
+        }
+      }
+  else
+    printf (" No directory structure at %.5lX\n", addr);
+}
+
+static void
+#ifdef __FunctionProto__
+do_list (int argc, char **argv)
+#else
+do_list (argc, argv)
+     int argc;
+     char *argv;
+#endif
+{
+  word_20 addr ;
+  unsigned int n;
+  
+  if (argc == 2 )
+  {
+    str_to_upper (argv[1]);
+    if (decode_20 (&addr, argv[1]))
+      {
+         printf(" %.5lX : %s\n", addr,  str_nibbles(addr, 16));
+      }
+  }
+  if (argc == 3 )
+  {
+    str_to_upper (argv[1]);
+    str_to_upper (argv[2]);
+
+    if ( decode_20(&addr, argv[1] ) && decode_dec(&n, argv[2] ) )
+      {
+         if (n <= 0)
+           return;
+
+         for ( ; n >0 ; n--)
+           {
+             printf(" %.5lX : %s\n", addr,  str_nibbles(addr, 16));
+             addr += 16;
+           }
+      }
+  }
+
+return;
+
+
+}
+
 static void
 #ifdef __FunctionProto__
 do_load (int argc, char **argv)
@@ -984,6 +1402,10 @@ do_mode (argc, argv)
 	{
 	  disassembler_mode = CLASS_MNEMONICS;
 	}
+      else if (!strcmp ("SYSRPL", argv[1]))
+	{
+          disassembler_mode = SYSRPL_MNEMONICS;
+	}
       else
 	{
 	  printf ("Unknown disassembler mode %s. Try \"help\".\n", argv[1]);
@@ -1023,11 +1445,10 @@ set_reg (val, n, r)
 
   for (i = 0; i < n; i++)
     {
-#ifdef SIMPLE_64
-      r[i] = (unsigned char) ((val & (0xf << (4 * i))) >> (4 * i));
-#else
-      r[i] = (unsigned char) (((i < 8?val.lo:val.hi) & (0xf << (4 * i))) >> (4 * i));
-#endif
+      if (i < 8)
+	r[i] = (unsigned char) ((val.lo & (0xf << (4 * i))) >> (4 * i));
+      else
+	r[i] = (unsigned char) ((val.hi & (0xf << (4 * (i - 8)))) >> (4 * (i - 8)));
     }
 }
 
@@ -1063,11 +1484,7 @@ set_st (val)
   int i;
 
   for (i = 0; i < 16; i++)
-#ifdef SIMPLE_64
-    saturn.PSTAT[i] = (val & (1 << i)) ? 1 : 0;
-#else
     saturn.PSTAT[i] = (val.lo & (1 << i)) ? 1 : 0;
-#endif
 }
 
 static void
@@ -1120,16 +1537,6 @@ set_hst (val)
   saturn.SB = 0;
   saturn.SR = 0;
   saturn.MP = 0;
-#ifdef SIMPLE_64
-  if (val & 1)
-    saturn.XM = 1;
-  if (val & 2)
-    saturn.SB = 1;
-  if (val & 4)
-    saturn.SR = 1;
-  if (val & 8)
-    saturn.MP = 1;
-#else
   if (val.lo & 1)
     saturn.XM = 1;
   if (val.lo & 2)
@@ -1138,7 +1545,6 @@ set_hst (val)
     saturn.SR = 1;
   if (val.lo & 8)
     saturn.MP = 1;
-#endif
 }
 
 static void
@@ -1163,7 +1569,7 @@ dump_hst ()
 }
 
 static char *mctl_str_gx[] = {
-  "MMIO       ",
+  "MMIO       ", 
   "SysRAM     ",
   "Bank Switch",
   "Port 1     ",
@@ -1172,7 +1578,7 @@ static char *mctl_str_gx[] = {
 };
 
 static char *mctl_str_sx[] = {
-  "MMIO  ",
+  "MMIO  ", 
   "SysRAM",
   "Port 1",
   "Port 2",
@@ -1190,6 +1596,45 @@ do_ram (argc, argv)
 #endif
 {
   int i;
+  word_20 home, end_home, len; 
+  char    buf[65536];
+
+  home     = read_nibbles(0x704EA,5);
+  printf("%.5lX : Key Buffer\n", home);
+
+  home     = read_nibbles(0x7057e,5);
+  end_home = read_nibbles(0x70579,5);
+  printf("%.5lX <-> %.5lX : Stack\n",home, end_home);
+
+  home     = read_nibbles(0x70583,5);
+  printf("%.5lX : Local vars\n", home);
+
+  home     = read_nibbles(0x70588,5);
+  printf("%.5lX : internal loops\n",home);
+
+  home     = read_nibbles(0x7058D,5);
+  printf("%.5lX : menu Keys\n",home);
+
+  home = read_nibbles(0x70592,5);
+  end_home = read_nibbles(0x70597,5);
+  printf("%.5lX <-> %.5lX : Home Directory\n",home, end_home);
+
+  home = end_home;
+  while(read_nibbles(home,5) !=0)
+    {
+      decode_rpl_obj(home, &len, buf);
+      if (strlen(buf) > 63)
+        {
+          sprintf(&buf[60], "...");
+          buf[63] = '\0';
+        }
+      printf("%.5lX -> %.5lX : %s\n", home, home+len , buf);
+      home += len;
+    }
+  printf("\n");
+
+
+  
 
   for (i = 0; i < 5; i++)
     {
@@ -1200,13 +1645,16 @@ do_ram (argc, argv)
         if (i == 0)
           printf("configured to 0x%.5lx\n", saturn.mem_cntl[i].config[0]);
         else
-          printf("configured to 0x%.5lX - 0x%.5lX\n",
+          {
+            printf("configured to 0x%.5lX - 0x%.5lX\n",
                  saturn.mem_cntl[i].config[0],
                  (saturn.mem_cntl[i].config[0] | ~saturn.mem_cntl[i].config[1])
                  & 0xfffff);
+          }
     }
   if (opt_gx)
     printf("Port 2      switched to bank %d\n", saturn.bank_switch);
+
 }
 
 static void
@@ -1387,11 +1835,7 @@ do_regs (argc, argv)
 	    }
 	  else if (!strcmp ("D0", argv[1]))
 	    {
-#ifdef SIMPLE_64
-	      saturn.D0 = (word_20)(val & 0xfffff);
-#else
 	      saturn.D0 = (word_20)(val.lo & 0xfffff);
-#endif
 	      printf ("    D0:\t%.5lX ->", saturn.D0);
 	      for (i = 0; i < 20; i += 5)
 		{
@@ -1401,11 +1845,7 @@ do_regs (argc, argv)
 	    }
 	  else if (!strcmp ("D1", argv[1]))
 	    {
-#ifdef SIMPLE_64
-	      saturn.D1 = (word_20)(val & 0xfffff);
-#else
 	      saturn.D1 = (word_20)(val.lo & 0xfffff);
-#endif
 	      printf ("    D1:\t%.5lX ->", saturn.D1);
 	      for (i = 0; i < 20; i += 5)
 		{
@@ -1415,20 +1855,12 @@ do_regs (argc, argv)
 	    }
 	  else if (!strcmp ("P", argv[1]))
 	    {
-#ifdef SIMPLE_64
-	      saturn.P = (word_4)(val & 0xf);
-#else
 	      saturn.P = (word_4)(val.lo & 0xf);
-#endif
 	      printf ("     P:\t%.1X\n", saturn.P);
 	    }
 	  else if (!strcmp ("PC", argv[1]))
 	    {
-#ifdef SIMPLE_64
-	      saturn.PC = (word_20)(val & 0xfffff);
-#else
 	      saturn.PC = (word_20)(val.lo & 0xfffff);
-#endif
 	      disassemble (saturn.PC, instr);
 	      printf ("    PC:\t%.5lX -> %s\n", saturn.PC, instr);
 	    }
@@ -1469,20 +1901,12 @@ do_regs (argc, argv)
 	    }
 	  else if (!strcmp ("CARRY", argv[1]))
 	    {
-#ifdef SIMPLE_64
-	      saturn.CARRY = (word_1)(val & 0x1);
-#else
 	      saturn.CARRY = (word_1)(val.lo & 0x1);
-#endif
 	      printf (" CARRY:\t%.1d\n", saturn.CARRY);
 	    }
 	  else if (!strcmp ("CY", argv[1]))
 	    {
-#ifdef SIMPLE_64
-	      saturn.CARRY = (word_1)(val & 0x1);
-#else
 	      saturn.CARRY = (word_1)(val.lo & 0x1);
-#endif
 	      printf (" CARRY:\t%.1d\n", saturn.CARRY);
 	    }
 	  else if (!strcmp ("ST", argv[1]))
@@ -1503,6 +1927,157 @@ do_regs (argc, argv)
     }
 }
 
+static void
+#ifdef __FunctionProto__
+do_receive (int argc, char **argv)
+#else
+do_receive (argc, argv)
+     int argc;
+     char *argv;
+#endif
+{
+   struct stat st;
+   FILE *fp;
+   unsigned char in[16] ;
+   char path[1024];
+   word_20 k, crc=0, addr, len, start=0xC0000;
+   int i, j;
+
+  if (argc == 2 )
+  {
+     get_home_directory(path);
+     strcat(path, "/");
+     strcat(path, argv[1]);
+
+     if ( NULL == (fp = fopen(path, "r")))
+       {
+         fprintf(stderr, "can\'t open file\n");
+         return ;
+       }
+
+     if (stat(path, &st) < 0)
+      {
+        fprintf(stderr, "can\'t stat \n");
+        fclose(fp);
+        return ;
+      }
+
+     if (fread(in, 1, 6, fp) != 6)
+      {
+        fprintf(stderr, "can\'t read first 6 bytes of \n");
+        fclose(fp);
+        return ;
+      }
+
+     if( strncmp(in,"HPHP48",6) )
+     {
+        printf("File is not a HP file\n");
+        fclose(fp);
+        return;
+     }
+     fseek(fp, 2, SEEK_CUR);
+
+  
+     addr = start;
+     write_nibbles(addr, 0x02b62 , 5); /* DOBACKUP */
+     addr +=5;
+     write_nibbles(addr, 0x0, 5);  /* len */
+     addr +=5; 
+
+     k = strlen(argv[1]);
+     write_nibbles(addr, k , 2); /* len of name */
+     addr +=2;
+ 
+     for (i=0; i < k; i++)
+       {
+         write_nibbles(addr, argv[1][i], 2);
+         addr+=2;
+       }
+
+     write_nibbles(addr, k , 2); /* len of name */
+     addr +=2;
+ 
+     printf("Contents of file %s:\n",path);
+     in[1] = 0;
+     k = 0;
+     while(fread(&in, 1, 1, fp))
+       {
+         i=(int)in[0] & 0xf0;
+         i >>= 4;
+         j = (int)in[0] & 0x0f;
+         j <<= 4;
+
+         printf( "%.2x", i+j );
+         write_nibbles(addr, in[0] , 2);
+         addr+= 2;
+         k++;
+       }
+     fclose(fp);
+     
+     len = addr - start +5 ;
+     write_nibbles(start+5, len, 5);
+     write_nibbles(addr, 0x2911, 5);
+     addr+=5;
+
+     addr+=1;
+     start += 5;
+     for ( ; start < addr ; start++)
+       crc = (crc >> 4) ^ (((crc ^ read_nibble(start)) & 0xf) * 0x1081);
+
+     write_nibbles(addr, crc, 5);
+     printf ("\n 0x%lxh nibbles, 0x%lxh CRC\n",len, crc);
+   }
+}
+
+static void
+#ifdef __FunctionProto__
+do_send (int argc, char **argv)
+#else
+do_send (argc, argv)
+     int argc;
+     char *argv;
+#endif
+{
+   word_20 addr, length , i;
+   struct stat st;
+   FILE *fp;
+   char byte[1], buf[65536], path[1024];
+  
+  if (argc == 2 )
+  {
+    if ((addr=where( argv[1])) != 0 )
+      {
+         get_home_directory(path);
+         strcat(path, "/");
+         strcat(path, argv[1]);
+         if ( NULL == (fp = fopen(path, "w")))
+           { 
+             fprintf(stderr, "can\'t open file : %s\n",path);
+             return ;
+           }
+        if (stat(path, &st) < 0)
+         {
+           fprintf(stderr, "can\'t stat file : %s\n",path);
+           fclose(fp);
+           return ;
+         }
+        decode_rpl_obj(addr, &length,  buf);
+        fwrite("HPHP48-E",1, 8, fp);
+        for( i =0 ; i <= length ; i+=2)
+          {
+            if( (length - i) == 1 )
+              byte[0] = (int) read_nibbles(addr,1);
+            else
+              byte[0] = (int)read_nibbles(addr,2);
+            fwrite( byte, 1, 1, fp);
+            addr += 2;
+          }
+        fclose(fp);
+        printf(" Wrote file : %s, 0x%lXh nibbles\n", path, length);
+      }
+  }
+}
+
 static void
 #ifdef __FunctionProto__
 do_save (int argc, char **argv)
@@ -1528,66 +2103,6 @@ struct se {
   struct se *se_next;
 };
 
-char *
-#ifdef __FunctionProto__
-get_stack (void)
-#else
-get_stack ()
-#endif
-{
-  word_20 dsktop, dskbot;
-  word_20 sp = 0, end = 0, ent = 0;
-  word_20 ram_base, ram_mask;
-  char    dat[65536];
-  char    typ[256];
-  int i, n;
-
-  ram_base = saturn.mem_cntl[1].config[0];
-  ram_mask = saturn.mem_cntl[1].config[1];
-  if (opt_gx)
-    {
-      saturn.mem_cntl[1].config[0] = 0x80000;
-      saturn.mem_cntl[1].config[1] = 0xc0000;
-      dsktop = DSKTOP_GX;
-      dskbot = DSKBOT_GX;
-    }
-  else
-    {
-      saturn.mem_cntl[1].config[0] = 0x70000;
-      saturn.mem_cntl[1].config[1] = 0xf0000;
-      dsktop = DSKTOP_SX;
-      dskbot = DSKBOT_SX;
-    }
-
-  load_addr(&sp, dsktop, 5);
-  load_addr(&end, dskbot, 5);
-
-  n = (end - sp) / 5 - 1;   /* end never matches sp */
-
-  printf("n = %d\n", n);
-
-  //TODO Get the top of the stack in the buffer.
-  //TODO create a window and put the complete stack into it!
-  if (n)
-    {
-      load_addr(&ent, sp, 5);
-      decode_rpl_obj_2(ent, typ, dat);
-      printf("%d %p -> [%s] %s\n", i, ent, typ, dat);
-    }
-
-  for (i = 0; i < n; i++)
-    {
-      load_addr(&ent, sp + (5 * i), 5);
-      decode_rpl_obj_2(ent, typ, dat);
-      printf("%d %p -> [%s] %s\n", i, ent, typ, dat);
-    }
-
-  saturn.mem_cntl[1].config[0] = ram_base;
-  saturn.mem_cntl[1].config[1] = ram_mask;
-
-  return;
-}
-
 static void
 #ifdef __FunctionProto__
 do_stack (int argc, char **argv)
@@ -1597,7 +2112,7 @@ do_stack (argc, argv)
      char *argv;
 #endif
 {
-  word_20 dsktop, dskbot;
+  word_20 dsktop, dskbot, len;
   word_20 sp = 0, end = 0, ent = 0;
   word_20 ram_base, ram_mask;
   char    buf[65536];
@@ -1652,7 +2167,7 @@ do_stack (argc, argv)
   se = stack;
   while (se)
     {
-      decode_rpl_obj(se->se_p, buf);
+      decode_rpl_obj(se->se_p, &len, buf);
       if (se->se_n != 1)
         if (strlen(buf) > 63)
           {
@@ -1661,7 +2176,7 @@ do_stack (argc, argv)
           }
       printf("%5d: %.5lX -> %s\n", se->se_n, se->se_p, buf);
       se = se->se_next;
-    }
+    } 
 
   se = stack;
   while (se)
@@ -1699,7 +2214,7 @@ do_step (argc, argv)
 #endif
 {
   word_20 next_instr;
-  word_32 n;
+  int n;
   int leave;
 
   if (enter_debugger & ILLEGAL_INSTRUCTION)
@@ -1710,13 +2225,20 @@ do_step (argc, argv)
 
   n = 1;
   if (argc > 1)
-    if (!decode_32 (&n, argv[1]))
+    if (!decode_dec(&n, argv[1]))
       return;
 
   if (n <= 0)
     return;
 
   in_debugger = 1;
+
+    if (check_breakpoint (BP_EXEC, saturn.PC) )
+    {
+       enter_debugger |= BREAKPOINT_HIT;
+       return;
+    }
+
   step_instruction ();
 
   if (exec_flags & EXEC_BKPT)
@@ -1748,6 +2270,12 @@ do_step (argc, argv)
 	  if (n == 0)
 	    break;
 	}
+    if (check_breakpoint (BP_EXEC, saturn.PC) )
+    {
+       enter_debugger |= BREAKPOINT_HIT;
+       return;
+    }
+
 
       step_instruction ();
 
@@ -1769,9 +2297,110 @@ do_step (argc, argv)
 
 static void
 #ifdef __FunctionProto__
-do_reset (int argc, char **argv)
+do_trace (int argc, char **argv)
 #else
-do_reset (argc, argv)
+do_step (argc, argv)
+     int argc;
+     char *argv;
+#endif
+{
+  word_20 next_instr;
+  unsigned int n;
+  int leave ;
+
+  if (enter_debugger & ILLEGAL_INSTRUCTION)
+    {
+      printf ("Can\'t step into an illegal instruction.");
+      return;
+    }
+
+  n = 1;
+  if (argc > 1)
+    if (!decode_dec(&n, argv[1]))
+      return;
+
+  if (n == 0)
+    return;
+
+  if ( n<0 )
+    {
+      leave = 1;
+      while (n != 0)
+         {
+           disassemble (saturn.old_PC[saturn.old_index], instr);
+           printf ("    PC:\t%.5lX -> %s\n", saturn.old_PC[saturn.old_index], instr);
+           if (saturn.old_index == 0)
+             saturn.old_index = 51;
+           saturn.old_index--;
+           n++;
+         }
+    }
+  else
+  {
+
+  in_debugger = 1;
+
+  step_instruction ();
+
+      /*
+       * print current instruction
+       */
+   disassemble (saturn.PC, instr);
+
+   if (check_breakpoint (BP_EXEC, saturn.PC))
+     printf("BREAK: ");
+   if( strncmp(instr, "nope",4) != 0 )
+     printf ("%.5lX: %s\n", saturn.PC, instr);
+   else
+     n++;
+
+  next_instr = saturn.PC;
+
+  sched_adjtime = 0;
+  schedule ();
+
+  enter_debugger = 0;
+  while (1)
+    {
+      if (enter_debugger)
+        break;
+
+      leave = 0;
+
+      if (saturn.PC == next_instr)
+        {
+          n--;
+          leave = 1;
+          if (n == 0)
+            break;
+        }
+
+    if (check_breakpoint (BP_EXEC, saturn.PC) )
+      printf("BREAK: ");
+
+    step_instruction ();
+    /*
+     * print current instruction
+     */
+    disassemble (saturn.PC, instr);
+
+    if( strncmp(instr, "nope",4) !=0 )
+      printf ("%.5lX: %s\n", saturn.PC, instr);
+    else
+      n++; 
+
+      if (leave)
+        next_instr = saturn.PC;
+      schedule ();
+    }
+  }
+
+}
+static void
+#ifdef __FunctionProto__
+do_reset_calc (int argc, char **argv)
+#else
+do_reset_calc (argc, argv)
      int argc;
      char *argv;
 #endif
@@ -1844,7 +2473,6 @@ debug ()
       if (enter_debugger & USER_INTERRUPT)
         if (verbose)
           printf ("%s: user interrupt (SIGINT) ignored\n", progname);
-        exit_x48(1);
       if (enter_debugger & BREAKPOINT_HIT)
         if (verbose)
           printf ("%s: breakpoint hit at 0x%.5lX ignored\n",
@@ -1897,13 +2525,22 @@ debug ()
       /*
        * print current instruction
        */
+/*
       disassemble (saturn.PC, instr);
       printf ("%.5lX: %s\n", saturn.PC, instr);
-
+*/
       /*
        * read a command
        */
+#ifdef HAVE_READLINE
       rl = readline ("x48-debug> ");
+#else
+      if (rl == (char *) 0)
+        rl = (char *)malloc((size_t)80);
+      printf("x48-debug> ");
+      fflush(stdout);
+      rl = read_str(rl, 80, 0);
+#endif
 
       if (rl == (char *) 0)
 	{
@@ -1926,6 +2563,10 @@ debug ()
 	}
       else
 	{
+#ifndef HAVE_READLINE
+          if (rl[strlen(rl) - 1] == '\n')
+            rl[strlen(rl) - 1] = '\0';
+#endif
 	  if (cl)
 	    {
 	      free (cl);
@@ -1990,7 +2631,7 @@ debug ()
   in_debugger = 1;
   ticks = get_t1_t2 ();
   in_debugger = 0;
-
+  
   if (saturn.t2_ctrl & 0x01)
     {
       saturn.timer2 = ticks.t2_ticks;
@@ -2039,6 +2680,11 @@ emulate_debug ()
 {
   do
     {
+    if (check_breakpoint (BP_EXEC, saturn.PC) )
+    {
+       enter_debugger |= BREAKPOINT_HIT;
+       return 0;
+    }
 
       step_instruction ();
 
