@@ -55,11 +55,9 @@
 #include "x48_x11.h"
 #include "debugger.h"
 
-#if 0
-#define DEBUG_TIMER
-#define DEBUG_SCHED 1
-#define DEBUG_DISP_SCHED
-#endif
+/* #define DEBUG_TIMER 1 */
+/* #define DEBUG_SCHED 1 */
+/* #define DEBUG_DISP_SCHED 1 */
 
 static long	jumpaddr;
 
@@ -122,6 +120,7 @@ decode_group_80()
   int t, op3, op4, op5, op6;
   unsigned char *REG;
   long addr;
+
   op3 = read_nibble(saturn.PC + 2);
   switch (op3) {
     case 0:	/* OUT=CS */
@@ -212,7 +211,7 @@ decode_group_80()
             REG = saturn.A;
           else
             REG = saturn.C;
-          if (op4 == 6 || op4 == 0xa)
+          if (op4 == 6 || op4 == 0xa) 
             t = 0;
           else
             t = 1;
@@ -704,6 +703,7 @@ int op1;
   int op2, op3, op4, op5, op6;
 
   op2 = read_nibble(saturn.PC + 1);
+
   switch (op1) {
     case 8:
       switch (op2) {
@@ -1975,7 +1975,13 @@ step_instruction()
 
   jumpaddr = 0;
 
+  if (saturn.old_index == 50 )
+    saturn.old_index = -1;
+  saturn.old_index++;
+  saturn.old_PC[saturn.old_index] = saturn.PC;
+
   op0 = read_nibble(saturn.PC);
+
   switch (op0) {
     case 0:
       op1 = read_nibble(saturn.PC + 1);
@@ -2206,6 +2212,7 @@ step_instruction()
       break;
   }
   instructions++;
+
   if (stop) {
     enter_debugger |= ILLEGAL_INSTRUCTION;
   }
@@ -2260,7 +2267,7 @@ schedule()
       if (device.display_touched < 0) device.display_touched = 1;
 #ifdef DEBUG_DISP_SCHED
       fprintf(stderr, "check_device: disp_when %d, disp_touched %d\n",
-              sched_display, device.display_touched);
+              device.disp_when, device.display_touched);
 #endif
     }
     check_devices();
@@ -2350,10 +2357,8 @@ schedule()
   if ((sched_statistics -= steps) <= 0) {
     sched_statistics = SCHED_STATISTICS;
     run = get_timer(RUN_TIMER);
-#ifndef SIMPLE_64
     s_1 = (run.hi << 19) | (run.lo >> 13);
     s_16 = (run.hi << 23) | (run.lo >> 9);
-#endif
     delta_t_1 = s_1 - old_s_1;
     delta_t_16 = s_16 - old_s_16;
     old_s_1 = s_1;
@@ -2428,19 +2433,13 @@ emulate(void)
 emulate()
 #endif
 {
-//unsigned char i;
-  struct timeval  tv;
-  struct timeval  tv2;
-#ifndef SOLARIS
-  struct timezone tz;
-#endif
-	
+
   reset_timer(T1_TIMER);
   reset_timer(RUN_TIMER);
   reset_timer(IDLE_TIMER);
 
-  set_accesstime();
   start_timer(T1_TIMER);
+  set_accesstime();
 
   start_timer(RUN_TIMER);
 
@@ -2450,35 +2449,17 @@ emulate()
   set_t1 = saturn.timer1;
 
   do {
+
     step_instruction();
 
-    //A short wait here! better N900 usage! :P
-    //for(i=0;i<2;i++);
-    
-    
-#ifdef SOLARIS
-    gettimeofday(&tv);
-#else
-    gettimeofday(&tv, &tz);
-#endif
-    while ((tv.tv_sec == tv2.tv_sec) && ((tv.tv_usec - tv2.tv_usec) < 0)) {
-	gettimeofday(&tv, &tz);
-    }
-
-    tv2.tv_usec = tv.tv_usec;
-    tv2.tv_sec = tv.tv_sec;
-
-/* We need to throttle the speed here. */
-
-//    if (schedule_event < 0) {
-//puts("bug");
-//	schedule_event = 0;
-//    }
-    if (schedule_event-- <= 0)
+    if (schedule_event-- == 0)
       {
         schedule();
       }
-  } while (!enter_debugger);
+
+/*  } while ( 0);   getz  */
+  } while (!enter_debugger); 
+
 
   return 0;
 }
