@@ -85,6 +85,7 @@
 #include "version.h"
 #include "resources.h"
 #include "romio.h"
+#include "entriesio.h"
 
 #define X48_MAGIC 0x48503438
 #define NR_CONFIG 8
@@ -269,6 +270,7 @@ init_saturn()
 
   memset(&saturn, 0, sizeof(saturn) - 4 * sizeof(unsigned char *));
   saturn.PC = 0x00000;
+  saturn.old_index = 50;
   saturn.magic = X48_MAGIC;
   saturn.t1_tick = 8192;
   saturn.t2_tick = 16;
@@ -1097,6 +1099,52 @@ const char *fname;
   return 1;
 }
 
+int
+#ifdef __FunctionProto__
+read_entries(const char *fname)
+#else
+read_entries(fname)
+const char *fname;
+#endif
+{
+  int ram_size;
+
+/*  if (!read_entries_file(fname, &saturn.entries))
+    return 0;
+*/
+
+  dev_memory_init();
+
+  if (opt_gx)
+    ram_size = RAM_SIZE_GX;
+  else
+    ram_size = RAM_SIZE_SX;
+
+  if (NULL == (saturn.ram = (word_4 *)malloc(ram_size)))
+    {
+      if (!quiet)
+        fprintf(stderr, "%s: can\'t malloc RAM\n", progname);
+      return 0;
+    }
+  memset(saturn.ram, 0, ram_size);
+
+  port1_size = 0;
+  port1_mask = 0;
+  port1_is_ram = 0;
+  saturn.port1 = (unsigned char *)0;
+
+  port2_size = 0;
+  port2_mask = 0;
+  port2_is_ram = 0;
+  saturn.port2 = (unsigned char *)0;
+
+  saturn.card_status = 0;
+
+  return 1;
+}
+
+
+
 void
 #ifdef __FunctionProto__
 get_home_directory(char *path)
@@ -1108,7 +1156,7 @@ char *path;
   char          *p;
   struct passwd *pwd;
 
-  if (homeDirectory[0] == '/')
+  if (homeDirectory[0] == (int)'/')
     {
       strcpy(path, homeDirectory);
     }
@@ -1156,15 +1204,18 @@ read_files()
   struct stat    st;
   FILE          *fp;
 
+  fprintf(stderr, "testing: Getting home dir\n");
   get_home_directory(path);
   strcat(path, "/");
 
   saturn.rom = (word_4 *)NULL;
+
+  fprintf(stderr, "testing: Reading ROM files\n");
   strcpy(fnam, path);
   strcat(fnam, "rom");
   if (!read_rom_file(fnam, &saturn.rom, &rom_size))
     return 0;
-
+ 
   rom_is_new = 0;
 
   strcpy(fnam, path);
@@ -1316,6 +1367,8 @@ read_files()
       exit (1);
     }
 
+  fprintf(stderr, "testing: Reading RAM file\n");
+
   strcpy(fnam, path);
   strcat(fnam, "ram");
   if ((fp = fopen(fnam, "r")) == NULL) {
@@ -1413,6 +1466,13 @@ read_files()
       saturn.card_status |= port2_is_ram ? 8 : 0;
     }
 
+  fprintf(stderr, "testing: Reading Enteries file\n");
+  strcpy(fnam, path);
+  strcat(fnam, "entries.o");
+  if (!read_entries_file(fnam, &saturn.entries)) 
+    return 0;
+
+
   return 1;
 }
 
@@ -1780,15 +1840,21 @@ init_emulator()
 #endif
 {
   if (!initialize)
+  {
+    fprintf(stderr, "testing: Read Files\n");
     if (read_files())
       {
         if (resetOnStartup)
           saturn.PC = 0x00000;
         return 0;
       }
-
+  }
+  fprintf(stderr, "Done Reading\n");
   init_saturn();
-  if (!read_rom(romFileName))
+  if (!read_rom(romFileName) )
+    exit(1); 
+
+  if (!read_entries(entriesFileName))
     exit(1);
   
   return 0;
