@@ -41,7 +41,7 @@
  * the above copyright notice appear in all copies and that both that
  * copyright notice and this permission notice appear in supporting
  * documentation.  No representations are made about the suitability of this
- * software for any purpose.  It is provided "as is" without express or
+ * software for any purpose.  It is provided "as is" without express or 
  * implied warranty.
  */
 
@@ -68,10 +68,10 @@ int     useSerial;
 char   *serialLine;
 int     useXShm;
 int     useDebugger;
-int	netbook;
 int     initialize;
 int     resetOnStartup;
 char   *romFileName;
+char   *entriesFileName;
 char   *homeDirectory;
 
 void
@@ -102,15 +102,12 @@ get_resources()
   resetOnStartup = get_boolean_resource("resetOnStartup",
                                         "ResetOnStartup");
   romFileName = get_string_resource("romFileName", "RomFileName");
+  entriesFileName = get_string_resource("enteriesFileName", "EnteriesFileName");
   homeDirectory = get_string_resource("homeDirectory", "HomeDirectory");
 
   useDebugger = get_boolean_resource("useDebugger", "UseDebugger");
   disassembler_mode = get_mnemonic_resource("disassemblerMnemonics",
                                             "DisassemblerMnemonics");
-
-  //netbook = get_boolean_resource("netbook", "Netbook");
-  //Only netbook mode for tablets
-  netbook = True;
 }
 
 char *
@@ -155,7 +152,7 @@ char *class;
   return get_string_resource_from_db(rdb, name, class);
 }
 
-int
+int 
 #ifdef __FunctionProto__
 get_mnemonic_resource (char *name, char *class)
 #else
@@ -183,7 +180,7 @@ char *class;
   return CLASS_MNEMONICS;
 }
 
-int
+int 
 #ifdef __FunctionProto__
 get_boolean_resource (char *name, char *class)
 #else
@@ -210,7 +207,7 @@ char *class;
   return 0;
 }
 
-int
+int 
 #ifdef __FunctionProto__
 get_integer_resource (char *name, char *class)
 #else
