@@ -47,11 +47,11 @@ extern int     useTerminal;
 extern int     useSerial;
 extern int     useXShm;
 extern int     useDebugger;
-extern int     netbook;
 extern char   *serialLine;
 extern int     initialize;
 extern int     resetOnStartup;
 extern char   *romFileName;
+extern char   *entriesFileName;
 extern char   *homeDirectory;
 
 extern char   *progname;
