@@ -39,7 +39,6 @@
 #include <stdio.h>
 #include <stdlib.h>
 #include <unistd.h>
-#include <string.h>
 
 #include "hp48.h"
 #include "hp48_emu.h"
@@ -48,6 +47,7 @@
 #include "append.h"
 #include "disasm.h"
 #include "romio.h"
+#include "entriesio.h"
 #define DEFINE_TRANS_TABLE 1
 #include "hp48char.h"
 #undef DEFINE_TRANS_TABLE
@@ -69,7 +69,7 @@ struct objfunc {
   { "String",		2,	DOCSTR,		dec_string },
   { "Hex String",	1,	DOHSTR,		dec_hex_string },
   { "List",		0,	DOLIST,		dec_list },
-  { "Directory",	0,	DORRP,		skip_ob },
+  { "Directory",	0,	DORRP,		dec_directory },
   { "Symbolic",		0,	DOSYMB,		dec_symb },
   { "Unit",		0,	DOEXT,		dec_unit },
   { "Tagged",		0,	DOTAG,		skip_ob },
@@ -146,6 +146,47 @@ long hxs;
   return n;
 }
 
+char *
+#ifdef __FunctionProto__
+dec_directory (word_20 *addr, char *string)
+#else
+dec_directory (addr, string)
+word_20 *addr;
+char    *string;
+#endif
+{
+  char    *p = string, buf[65536];
+  word_20  start, len;
+  int      home;
+
+  if (*addr == read_nibbles(0x70592,5))
+    home = 1;
+  else
+    home =0;
+
+  start = read_nibbles(*addr,3);
+  *addr += 3;
+  if ( home )
+    {
+      for ( ; start >= 1 ; start-- )
+      *addr+=13;
+    }
+   *addr += read_nibbles(*addr,5) ;
+   if ( read_nibbles(*addr,5) != 0 )
+     {
+       len = read_nibbles(*addr, 2);
+       if ( len == 0 )
+         *addr += 2;
+       else
+         *addr += (len+2)*2;
+     }
+
+  decode_rpl_obj(*addr, &len, buf);
+  *addr += len;
+
+  return p;
+}
+
 char *
 #ifdef __FunctionProto__
 dec_bin_int(word_20 *addr, char *string)
@@ -204,7 +245,7 @@ int      xl;
   while (--xl) xs *= 10;
   re = hxs2real(r.x);
   if (re >= xs)
-    re = re - 2 * xs;
+    re = re - 2 * xs; 
 
 
   if ((re >= 0) && (re < ml + 1))
@@ -482,7 +523,7 @@ char    *string;
   *p++ = '}';
   *p = '\0';
 
-  *addr += 5;
+  *addr += 3;  /* was 5 */
   return p;
 }
 
@@ -666,10 +707,10 @@ char    *string;
       size = 10;
       sprintf(p, "ACPTR ");
       p += strlen(p);
-      for (i = 0; i < 5; i++)
+      for (i = 0; i < 5; i++) 
         *p++ = hex[read_nibble(*addr + i)];
       *p++ = ' ';
-      for (i = 5; i < 10; i++)
+      for (i = 5; i < 10; i++) 
         *p++ = hex[read_nibble(*addr + i)];
     }
   else
@@ -1405,93 +1446,24 @@ char    *string;
   return p;
 }
 
-void
-#ifdef __FunctionProto__
-decode_rpl_obj_2(word_20 addr, char *typ, char *dat)
-#else
-decode_rpl_obj_2(addr, typ, dat)
-word_20  addr;
-char    *typ;
-char    *dat;
-#endif
-{
-  word_20	  prolog = 0;
-  int             len;
-  char		  tmp_str[80];
-  struct objfunc *op;
-
-  typ[0] = '\0';
-  dat[0] = '\0';
-
-  prolog = read_nibbles(addr, 5);
-
-  for (op = objects; op->prolog != 0; op++)
-    {
-      if (op->prolog == prolog)
-        break;
-    }
-
-  if (op->prolog == 0)
-    {
-      if (addr == SEMI)
-        {
-          append_str(typ, "Primitive Code");
-          append_str(dat, "SEMI");
-        }
-      else if (addr + 5 == prolog)
-        {
-          append_str(typ, "Primitive Code");
-          sprintf(dat, "at %.5lX", prolog);
-        }
-      else
-        {
-          append_str(typ, "PTR");
-          sprintf(dat, "%.5lX", prolog);
-        }
-      return;
-    }
-
-  if (op->prolog == DOCOL)
-    {
-      if (check_xlib(addr, tmp_str))
-        {
-          append_str(typ, "XLib Call");
-          append_str(dat, tmp_str);
-	  return;
-        }
-    }
-
-  if (op->length)
-    {
-      len = (read_nibbles(addr + 5, 5) - 5) / op->length;
-      sprintf(typ, "%s %d", op->name, len);
-    }
-  else
-    {
-  append_str(typ, op->name);
-    }
-
-  addr += 5;
-  (*op->func)(&addr, dat);
-
-  return;
-}
-
 char *
 #ifdef __FunctionProto__
-decode_rpl_obj(word_20 addr, char *buf)
+decode_rpl_obj(word_20 addr, word_20 *length, char *buf)
 #else
-decode_rpl_obj(addr, buf)
+decode_rpl_obj(addr, length, buf)
 word_20  addr;
+word_20 *length;
 char    *buf;
 #endif
 {
   word_20	  prolog = 0;
+  word_20         start;
   int             len;
   char           *p = buf;
   char		  tmp_str[80];
   struct objfunc *op;
 
+  start=addr;
   prolog = read_nibbles(addr, 5);
 
   for (op = objects; op->prolog != 0; op++)
@@ -1551,6 +1523,7 @@ char    *buf;
   addr += 5;
   p = (*op->func)(&addr, p);
 
+  *length = addr-start;
   return p;
 }
 
